# -*- coding: utf-8 -*-

import re
from html import escape


CURRENT_VERSION = 1
CURRENT_VERSION_TEXT = "\x01"


class BBCodeParseError(Exception):
    def __init__(self, message):
        super(BBCodeParseError, self).__init__(message)


class BBCodeNode(object):
    html_start = ""
    html_end = ""
    bbcode_ident = ""
    final = False
    allow_param = False
    mandat_param = False
    auto_close = False
    allowed_children = None
    mandat_parent = None

    def __init__(self):
        self.sub_nodes = []

    def append(self, item):
        self.sub_nodes.append(item)

    def as_bbcode(self):
        bbcode = ""
        for item in self.sub_nodes:
            bbcode += item.as_bbcode()
        if self.bbcode_ident:
            bbcode = "[" + self.bbcode_ident + "]" + bbcode + "[/" + self.bbcode_ident + "]"
        return bbcode

    def as_html(self, **kwargs):
        html = self.html_start
        for item in self.sub_nodes:
            html += item.as_html(**kwargs)
        html += self.html_end
        return html

    def as_text(self):
        text = ""
        for item in self.sub_nodes:
            text += item.as_text()
        return text

    def get_attachments(self):
        attachments = []
        self._get_attachments_into(attachments)
        return attachments

    def _get_attachments_into(self, attachments):
        for item in self.sub_nodes:
            item._get_attachments_into(attachments)

    def post_validate(self):
        for item in self.sub_nodes:
            item.post_validate()

    def __repr__(self):
        if self.final:
            return f"<{self.__class__.__name__}>"
        return "<{name}: [{sub_nodes}]>".format(
            name=self.__class__.__name__,
            sub_nodes=",".join([repr(i) for i in self.sub_nodes]),
        )


class BBCodeRootNode(BBCodeNode):
    def __init__(self, version):
        super(BBCodeRootNode, self).__init__()
        self.version = version

    def serialize(self):
        return CURRENT_VERSION_TEXT + self.as_bbcode()


class BBCodeLocalNode(object):
    allow_param = False
    mandat_param = False
    allowed_children = None
    mandat_parent = None
    auto_close = False

    def as_bbcode(self):
        return "[" + self.bbcode_ident + "]"

    def as_html(self, **kwargs):
        return self.html_code

    def as_text(self):
        return ""

    def post_validate(self):
        pass

    def _get_attachments_into(self, attachments):
        pass


class BBCodeTextNode(BBCodeNode):
    def __init__(self, text):
        self.text = text.replace("\r\n", "\n")
        super(BBCodeTextNode, self).__init__()

    def as_bbcode(self):
        return self.text

    def as_html(self, **kwargs):
        text = self.text
        if text[0:1] == "\n":
            return "\n" + escape(text[1:], True).replace("\n", "<br />\n")

        if text[-1:] == "\n":
            return escape(text[:-1], True).replace("\n", "<br />\n") + "\n"

        return escape(text, True).replace("\n", "<br />\n")

    def as_text(self):
        return self.text

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self.text)}>"


class BBCodeFinalNode(BBCodeNode):
    final = True

    def __init__(self, text):
        super(BBCodeFinalNode, self).__init__()
        self.validate(text)
        self.text = text

    def as_bbcode(self):
        return "[" + self.bbcode_ident + "]" + self.text + "[/" + self.bbcode_ident + "]"

    def as_html(self, **kwargs):
        return self.html_start + escape(self.text) + self.html_end

    def validate(self, text):
        pass

    def post_validate(self):
        pass


class BBCodeParameterNode(BBCodeNode):
    allow_param = True

    def validate_all(self, parameter):
        raise NotImplementedError("Must be redefined in subclasses")

    def __init__(self, parameter=None):
        super(BBCodeParameterNode, self).__init__()
        if parameter is None:
            self.parameter = None
        else:
            self.parameter = parameter
        self.validate_all(parameter)

    def as_bbcode(self):
        if self.parameter:
            bbcode = "[" + self.bbcode_ident + "=" + self.parameter + "]"
        else:
            bbcode = "[" + self.bbcode_ident + "]"

        for item in self.sub_nodes:
            bbcode += item.as_bbcode()

        bbcode += "[/" + self.bbcode_ident + "]"

        return bbcode


class BBCodeFinalParameterNode(BBCodeFinalNode):
    allow_param = True

    def validate_all(self, text, parameter):
        raise NotImplementedError("Must be redefined in subclasses")

    def __init__(self, text, parameter):
        super(BBCodeFinalParameterNode, self).__init__(text)
        if parameter is None:
            self.parameter = None
        else:
            self.parameter = parameter
        self.validate_all(text, parameter)

    def as_bbcode(self):
        if self.parameter:
            return (
                "["
                + self.bbcode_ident
                + "="
                + escape(self.parameter)
                + "]"
                + self.text
                + "[/"
                + self.bbcode_ident
                + "]"
            )
        else:
            return "[" + self.bbcode_ident + "]" + self.text + "[/" + self.bbcode_ident + "]"


class BBCodeParagraphNode(BBCodeParameterNode):
    html_start = "<p>"
    html_end = "</p>"
    bbcode_ident = "p"

    def validate_all(self, parameter):
        if parameter and parameter not in ["left", "right", "center", "justify"]:
            raise BBCodeParseError("Paramètre d’alignement incorrect")

    def as_html(self, **kwargs):
        if self.parameter:
            html = '<p class="align' + self.parameter.capitalize() + '">'
        else:
            html = "<p>"
        for item in self.sub_nodes:
            html += item.as_html(**kwargs)
        html += self.html_end
        return html

    def as_text(self):
        text = "\n"
        for item in self.sub_nodes:
            text += item.as_text()
        text += "\n"
        return text


class BBCodeBoldNode(BBCodeNode):
    html_start = "<b>"
    html_end = "</b>"
    bbcode_ident = "b"


class BBCodeItalicNode(BBCodeNode):
    html_start = "<i>"
    html_end = "</i>"
    bbcode_ident = "i"


class BBCodeUnderlineNode(BBCodeNode):
    html_start = '<span style="text-decoration:underline">'
    html_end = "</span>"
    bbcode_ident = "u"


class BBCodeStrikeNode(BBCodeNode):
    """Strike Text"""

    html_start = "<del>"
    html_end = "</del>"
    bbcode_ident = "s"


class BBCodeQuoteNode(BBCodeParameterNode):
    bbcode_ident = "quote"

    def validate_all(self, parameter):
        # On autorise tout paramètre
        return

    def as_html(self, **kwargs):
        s = '<div class="quote"><span class="quote">❝</span> '
        if self.parameter:
            s += '<span class="quoted">' + escape(self.parameter) + "</span>"
        s += "<br />\n"

        s += super(BBCodeQuoteNode, self).as_html(**kwargs)

        s += "</div>"

        return s

    def as_text(self):
        text = "\n« "
        for item in self.sub_nodes:
            text += item.as_text()
        text += " »\n"
        return text


class BBCodeCodeNode(BBCodeFinalNode):
    html_start = "<pre>"
    html_end = "</pre>"
    bbcode_ident = "code"

    def as_text(self):
        return "\n[Code]\n"


class BBCodeListNode(BBCodeNode):
    allowed_children = ["BBCodeListItemNode"]
    html_start = "<ul>"
    html_end = "</ul>"
    bbcode_ident = "list"

    def as_text(self):
        text = "\n"
        for item in self.sub_nodes:
            text += ("    • " if isinstance(item, BBCodeListItemNode) else "") + item.as_text() + "\n"
        return text


class BBCodeListItemNode(BBCodeNode):
    mandat_parent = BBCodeListNode
    html_start = "<li>"
    html_end = "</li>"
    bbcode_ident = "*"
    auto_close = True


url_re = re.compile(
    r"^(https?://|mailto: *|/)([a-zA-Z0-9$-/_=:;!*\(\),~@]|(%[0-9a-fA-F][0-9a-fA-F]))+(\?([^#]*))?(#(.*))?$"
)


class BBCodeURLNode(BBCodeParameterNode):
    bbcode_ident = "url"
    allowed_children = ["BBCodeImgNode", "BBCodeBoldNode"]

    def validate_all(self, parameter):
        # Paramètre validé dans post_validate
        return

    def _inner_text(self):
        if len(self.sub_nodes) == 1 and isinstance(self.sub_nodes[0], BBCodeTextNode):
            return self.sub_nodes[0].text

        return ""

    def as_html(self, **kwargs):
        if self.parameter:
            inner_contents = super(BBCodeURLNode, self).as_html(**kwargs)
            return '<a href="' + escape(self.parameter, True) + '">' + inner_contents + "</a>"
        else:
            return '<a href="' + escape(self._inner_text(), True) + '">' + escape(self._inner_text()) + "</a>"

    def as_text(self):
        if self.parameter:
            return super(BBCodeURLNode, self).as_text()

        return self._inner_text()

    def post_validate(self):
        super(BBCodeURLNode, self).post_validate()
        if self.parameter:
            url = self.parameter
        else:
            url = self._inner_text()

        if not url_re.match(url):
            raise BBCodeParseError("URL " + url + " invalide.")


class BBCodeSizeNode(BBCodeParameterNode):
    mandat_param = True
    bbcode_ident = "size"

    def validate_all(self, parameter):
        try:
            size = int(parameter)
        except ValueError:
            size = 0

        if size < 50 or size > 200:
            raise BBCodeParseError("Taille de texte invalide.")

    def as_html(self, **kwargs):
        s = f'<span style="font-size:{self.parameter}%">'

        s += super(BBCodeSizeNode, self).as_html(**kwargs)

        s += "</span>"
        return s


class BBCodeColorNode(BBCodeParameterNode):
    mandat_param = True
    bbcode_ident = "color"

    def validate_all(self, parameter):
        if not re.match(r"^#[0-9A-F]{6}$", parameter):
            raise BBCodeParseError("Couleur invalide.")

    def as_html(self, **kwargs):
        s = '<span style="color:' + self.parameter + '">'

        s += super(BBCodeColorNode, self).as_html(**kwargs)

        s += "</span>"
        return s


class BBCodeLineBreakNode(BBCodeLocalNode):
    bbcode_ident = "br"
    html_code = "<br />"


class BBCodeH1Node(BBCodeNode):
    html_start = "<h1>"
    html_end = "</h1>"
    bbcode_ident = "h1"


class BBCodeH2Node(BBCodeNode):
    html_start = "<h2>"
    html_end = "</h2>"
    bbcode_ident = "h2"


class BBCodeH3Node(BBCodeNode):
    html_start = "<h3>"
    html_end = "</h3>"
    bbcode_ident = "h3"


class BBCodeAttachmentNode(BBCodeFinalNode):
    bbcode_ident = "attach"

    def __init__(self, path):
        self.path = path

    def as_bbcode(self):
        return "[" + self.bbcode_ident + "]" + self.path + "[/" + self.bbcode_ident + "]"

    def as_html(self, attachment_callback, **kwargs):
        assert attachment_callback is not None
        return attachment_callback(self.path)

    def as_text(self):
        return ""

    def _get_attachments_into(self, attachments):
        if self.path not in attachments:
            attachments.append(self.path)


class BBCodeImgNode(BBCodeFinalParameterNode):
    bbcode_ident = "img"

    def validate_all(self, text, parameter):
        if not url_re.match(text):
            raise BBCodeParseError("URL d’image invalide.")

    def as_html(self, **kwargs):
        if self.parameter:
            return '<img src="' + escape(self.text, True) + '" alt="' + escape(self.parameter, True) + '" />'
        else:
            return '<img src="' + escape(self.text, True) + '" alt="" />'

    def as_text(self):
        return "" if self.parameter is None else self.parameter


BBCODE_PACKS = {
    "message": {
        "b": BBCodeBoldNode,
        "i": BBCodeItalicNode,
        "u": BBCodeUnderlineNode,
        "s": BBCodeStrikeNode,
        "quote": BBCodeQuoteNode,
        "code": BBCodeCodeNode,
        "list": BBCodeListNode,
        "*": BBCodeListItemNode,
        "url": BBCodeURLNode,
        "size": BBCodeSizeNode,
        "color": BBCodeColorNode,
        "br": BBCodeLineBreakNode,
        "img": BBCodeImgNode,
    },
    "web_page": {
        "p": BBCodeParagraphNode,
        "b": BBCodeBoldNode,
        "i": BBCodeItalicNode,
        "u": BBCodeUnderlineNode,
        "s": BBCodeStrikeNode,
        "quote": BBCodeQuoteNode,
        "code": BBCodeCodeNode,
        "list": BBCodeListNode,
        "*": BBCodeListItemNode,
        "url": BBCodeURLNode,
        "size": BBCodeSizeNode,
        "color": BBCodeColorNode,
        "br": BBCodeLineBreakNode,
        "img": BBCodeImgNode,
        "h1": BBCodeH1Node,
        "h2": BBCodeH2Node,
        "h3": BBCodeH3Node,
    },
    "message_attach": {
        "b": BBCodeBoldNode,
        "i": BBCodeItalicNode,
        "u": BBCodeUnderlineNode,
        "s": BBCodeStrikeNode,
        "quote": BBCodeQuoteNode,
        "code": BBCodeCodeNode,
        "list": BBCodeListNode,
        "*": BBCodeListItemNode,
        "url": BBCodeURLNode,
        "size": BBCodeSizeNode,
        "color": BBCodeColorNode,
        "br": BBCodeLineBreakNode,
        "img": BBCodeImgNode,
        "attach": BBCodeAttachmentNode,
    },
    "web_page_attach": {
        "p": BBCodeParagraphNode,
        "b": BBCodeBoldNode,
        "i": BBCodeItalicNode,
        "u": BBCodeUnderlineNode,
        "s": BBCodeStrikeNode,
        "quote": BBCodeQuoteNode,
        "code": BBCodeCodeNode,
        "list": BBCodeListNode,
        "*": BBCodeListItemNode,
        "url": BBCodeURLNode,
        "size": BBCodeSizeNode,
        "color": BBCodeColorNode,
        "br": BBCodeLineBreakNode,
        "img": BBCodeImgNode,
        "h1": BBCodeH1Node,
        "h2": BBCodeH2Node,
        "h3": BBCodeH3Node,
        "attach": BBCodeAttachmentNode,
    },
}

VALID_PACKS = {key: set(value.keys()) for (key, value) in BBCODE_PACKS.items()}


def get_valid_bbcode_packs():
    return VALID_PACKS


def _parse_bbcode(pack, version, text):
    try:
        defined_bbcodes = BBCODE_PACKS[pack]
    except KeyError:
        raise ValueError("Unknown BBCode pack " + pack)

    def new_node(name, *args):
        return defined_bbcodes[name](*args)

    curr = 0
    length = len(text)

    root_node = BBCodeRootNode(version=version)
    node_stack = [root_node]
    current_node = root_node

    while True:
        open = text.find("[", curr)

        if open == -1:
            if text[curr:]:
                current_node.append(BBCodeTextNode(text[curr:]))
            curr = length
            break

        new_pos = 0
        search = re.match(r"^\[([a-zA-Z1-9]+|\*)(=([^\]]+))?\]", text[open:])
        redo_loop = False
        if search:  # BBCode de début
            name = search.group(1).lower()
            if name in defined_bbcodes:
                if defined_bbcodes[name].auto_close and isinstance(current_node, defined_bbcodes[name]):
                    # Exemple : [*] blabla... [*]
                    # On suppose que le BBCode précédent est mal fermé
                    if open > curr:
                        current_node.append(BBCodeTextNode(text[curr:open]))
                        curr = open
                    node_stack.pop()
                    current_node = node_stack[-1]

                # Vérification de l’imbrication des BBCodes
                if (
                    current_node.allowed_children is not None
                    and defined_bbcodes[name].__name__ not in current_node.allowed_children
                ):
                    raise BBCodeParseError(
                        "Le BBCode [" + current_node.bbcode_ident + "] ne peut contenir de BBCode [" + name + "]."
                    )

                if (
                    defined_bbcodes[name].mandat_parent is not None
                    and defined_bbcodes[name].mandat_parent != current_node.__class__
                ):
                    erreur = "Le BBCode [{}] ne peut se trouver que dans un BBCode [{}].".format(
                        name, defined_bbcodes[name].mandat_parent.bbcode_ident
                    )
                    raise BBCodeParseError(erreur)

                # Vérification de la prise ou non de paramètres.
                if search.group(3):
                    if not defined_bbcodes[name].allow_param:
                        raise BBCodeParseError("Le BBCode [" + name + "] ne prend pas de paramètre.")
                else:
                    if defined_bbcodes[name].mandat_param:
                        raise BBCodeParseError("Le BBCode [" + name + "] nécessite un paramètre.")

                if issubclass(defined_bbcodes[name], BBCodeLocalNode):
                    # BBCode sans balise de fermeture ([br] par exemple)
                    node = new_node(name)
                    if open > curr:
                        current_node.append(BBCodeTextNode(text[curr:open]))
                    current_node.append(node)
                    new_pos = open + len(search.group(0))

                elif defined_bbcodes[name].final:
                    # BBCode traitant directement le code entre ses balises.
                    # ([url], [img], [code] par exemple)
                    end = text.find("[/{}]".format(name), open)
                    if end == -1:
                        raise BBCodeParseError("BBCode [{}] mal fermé.".format(name))
                    if defined_bbcodes[name].allow_param:
                        start_pos = open + len(search.group(1))
                        if search.group(2):
                            start_pos += len(search.group(2))
                        start_pos += 2
                        node = new_node(name, text[start_pos:end], search.group(3))
                    else:
                        node = new_node(name, text[open + len(search.group(1)) + 2 : end])
                    if open > curr:
                        current_node.append(BBCodeTextNode(text[curr:open]))
                    current_node.append(node)
                    new_pos = end + len(search.group(1)) + 3

                else:
                    # Type de BBCode par défaut
                    if defined_bbcodes[name].allow_param:
                        node = new_node(name, search.group(3))
                    else:
                        node = new_node(name)
                    if open > curr:
                        current_node.append(BBCodeTextNode(text[curr:open]))
                    current_node.append(node)
                    node_stack.append(node)
                    current_node = node
                    new_pos = open + len(search.group(0))
        else:
            search = re.match(r"^\[/([a-zA-Z1-9]+|\*)\]", text[open:])
            if search:  # BBCode de fin
                name = search.group(1).lower()
                if name in defined_bbcodes and not issubclass(defined_bbcodes[name], BBCodeLocalNode):
                    if open > curr:
                        current_node.append(BBCodeTextNode(text[curr:open]))
                    if isinstance(current_node, defined_bbcodes[name]):
                        # C’est bien le BBCode que l’on doit fermer
                        node_stack.pop()
                        current_node = node_stack[-1]
                        new_pos = open + len(search.group(1)) + 3
                    elif current_node.auto_close:
                        # On considère que le nœud courant est fermé automatiquement
                        node_stack.pop()
                        current_node = node_stack[-1]
                        new_pos = open
                        # ... et on refait un tour...
                        redo_loop = True
                    else:
                        raise BBCodeParseError("BBCode de fermeture [/{}] mal placé.".format(name))

        if new_pos:
            if curr >= new_pos:
                if curr > new_pos or not redo_loop:  # pragma: no cover
                    raise Exception("Infinite loop!")
            curr = new_pos
        else:
            new_pos = text.find("[", open + 1)
            if new_pos == -1:
                current_node.append(BBCodeTextNode(text[curr:]))
                break
            current_node.append(BBCodeTextNode(text[curr:new_pos]))
            if curr >= new_pos:
                raise Exception("Infinite loop!")  # pragma: no cover
            curr = new_pos

    # end while
    if len(node_stack) != 1:
        ident = node_stack[-1].bbcode_ident
        raise BBCodeParseError(f"BBCode [{ident}] mal fermé.")

    root_node.post_validate()

    return root_node


def deserialize(pack, text):
    version = ord(text[0]) if len(text) >= 1 else None
    if version != CURRENT_VERSION:
        raise BBCodeParseError("Version {} incorrecte".format(version))

    return _parse_bbcode(pack, version, text[1:])


def parse_user_input(pack, text):
    return _parse_bbcode(pack, CURRENT_VERSION, text)


def bbcode_from_text(text):
    root_node = BBCodeRootNode(version=CURRENT_VERSION)
    root_node.append(BBCodeTextNode(text))
    return root_node
